/*
 *    HelpViewer - Simple Help file browser
 *    Copyright (C) 2009 Leandro A. F. Pereira <leandro@hardinfo.org>
 *    Copyright (C) 2015 James B
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, version 2.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
 */

#define _GNU_SOURCE		/* for strcasestr() */
#include <string.h>
#include <stdlib.h>
#include <gtk/gtk.h>
#include <libintl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <locale.h>

#include "markdown-text-view.h"
#include "help-viewer.h"

#define DEFAULT_INDEX "index.md"
#define DEFAULT_WINDOW_TITLE "mdview"
#define MDVIEW_VERSION "2016.04.26"

# define MDVIEW_VERSION_TEXT "\
mdview " MDVIEW_VERSION " built on " __TIMESTAMP__ "\n\
(C) Leandro Pereira 2009, James B 2015, 2016, step 2016.\n\
"

#define _(STRING) gettext(STRING)
void open_url(gchar * url)
{
    const gchar *browsers[] = {
	"xdg-open", "gnome-open", "kfmclient openURL",
	"sensible-browser", "firefox", "epiphany",
	"iceweasel", "seamonkey", "galeon", "mozilla",
	"opera", "konqueror", "netscape", "links -g",
	NULL
    };
    gint i = 0;
    gchar *browser = (gchar *)g_getenv("BROWSER");
    
    if (!browser || *browser == '\0') {
    	browser = (gchar *)browsers[i++];
    }
    
    do {
		gchar *cmdline = g_strdup_printf("%s '%s'", browser, url);

		if (g_spawn_command_line_async(cmdline, NULL)) {
			g_free(cmdline);
			return;
		}

		g_free(cmdline);
    	browser = (gchar *)browsers[i++];
    	
    } while (browser);

    g_warning(_("Couldn't find a Web browser to open URL %s."), url);
}

static void do_file_search(HelpViewer *hv, gchar *text);

static void forward_clicked(GtkWidget *widget, gpointer data)
{
    HelpViewer *hv = (HelpViewer *)data;
    GSList *temp;
    
    /* puts the current file on the back stack */
    hv->back_stack = g_slist_prepend(hv->back_stack, g_strdup(hv->current_file));
    
    /* enables the back button */
    gtk_widget_set_sensitive(hv->btn_back, hv->can_go_back=TRUE);

    /* loads the new current file */
    if (g_str_has_prefix(hv->forward_stack->data, "search://")) {
        do_file_search(hv, hv->forward_stack->data + sizeof("search://") - 1);
    } else {
        markdown_textview_load_file(MARKDOWN_TEXTVIEW(hv->text_view), hv->forward_stack->data);
    }
    
    /* pops the stack */
    temp = hv->forward_stack->next;
    g_free(hv->forward_stack->data);
    g_slist_free1(hv->forward_stack);
    hv->forward_stack = temp;
    
    /* if there aren't items on forward stack anymore, disables the button */
    if (!hv->forward_stack) {
        gtk_widget_set_sensitive(hv->btn_forward, hv->can_go_forward=FALSE);
    }
}

static gboolean forward_key_pressed ( GtkAccelGroup  *group,
          GObject        *obj,
          guint           keyval,
          GdkModifierType mod,
          gpointer        data ) 
{
	HelpViewer *hv = (HelpViewer *)data;
	if (hv->can_go_forward)
		forward_clicked(NULL, data);
	return TRUE;
}

static void back_clicked(GtkWidget *widget, gpointer data)
{
    HelpViewer *hv = (HelpViewer *)data;
    GSList *temp;
    
    /* puts the current file on the forward stack */
    hv->forward_stack = g_slist_prepend(hv->forward_stack, g_strdup(hv->current_file));
    
    /* enables the forward button */
    gtk_widget_set_sensitive(hv->btn_forward, hv->can_go_forward=TRUE);

    /* loads the new current file */
    if (g_str_has_prefix(hv->back_stack->data, "search://")) {
        do_file_search(hv, hv->back_stack->data + sizeof("search://") - 1);
    } else {
        markdown_textview_load_file(MARKDOWN_TEXTVIEW(hv->text_view), hv->back_stack->data);
    }
    
    /* pops the stack */
    temp = hv->back_stack->next;
    g_free(hv->back_stack->data);
    g_slist_free1(hv->back_stack);
    hv->back_stack = temp;
    
    /* if there aren't items on back stack anymore, disables the button */
    if (!hv->back_stack) {
        gtk_widget_set_sensitive(hv->btn_back, hv->can_go_back=FALSE);
    }
}

static gboolean backward_key_pressed ( GtkAccelGroup  *group,
          GObject        *obj,
          guint           keyval,
          GdkModifierType mod,
          gpointer        data ) 
{
	HelpViewer *hv = (HelpViewer *)data;
	if (hv->can_go_back)
		back_clicked(NULL, data);
	return TRUE;
}

static void link_clicked(MarkdownTextView *text_view, gchar *link, gpointer data)
{
    HelpViewer *hv = (HelpViewer *)data;
    
    if (g_str_has_prefix(link, "http://") || 
        g_str_has_prefix(link, "https://") ||
        g_str_has_prefix(link, "ftp://")) {
        open_url(link);
    } else {
        /* adds the current file to the back stack (before loading the new file */
        hv->back_stack = g_slist_prepend(hv->back_stack, g_strdup(hv->current_file));
        gtk_widget_set_sensitive(hv->btn_back, hv->can_go_back=TRUE);

        gtk_statusbar_pop(GTK_STATUSBAR(hv->status_bar), 1);
        markdown_textview_load_file(text_view, link);
    }
}

static void file_load_complete(MarkdownTextView *text_view, gchar *file, gpointer data)
{
    HelpViewer *hv = (HelpViewer *)data;
    
    /* sets the currently-loaded file */
    g_free(hv->current_file);
    hv->current_file = g_strdup(file);
    
    gtk_statusbar_push(GTK_STATUSBAR(hv->status_bar), 1, "Done.");
    markdown_textview_cursor_to_top(MARKDOWN_TEXTVIEW(hv->text_view));
}

static void hovering_over_link(MarkdownTextView *text_view, gchar *link, gpointer data)
{
    HelpViewer *hv = (HelpViewer *)data;
    gchar *temp;
    
    temp = g_strconcat("Link to ", link, NULL);
    
    gtk_statusbar_push(GTK_STATUSBAR(hv->status_bar), 1, temp);
    
    g_free(temp);
}

static void hovering_over_text(MarkdownTextView *text_view, gpointer data)
{
    HelpViewer *hv = (HelpViewer *)data;
    
    gtk_statusbar_pop(GTK_STATUSBAR(hv->status_bar), 1);
}

static void do_file_search(HelpViewer *hv, gchar *text)
{
    GString *markdown;
    GDir *dir;
    gchar **terms;
    gint no_results = 0;
    int term;
    
    /*
     * FIXME: This search is currently pretty slow; think on a better way to do search. 
     * Ideas:
     * - Build a index the first time the help file is opened
     * - Search only titles and subtitles
     */
    
    terms = g_strsplit(text, " ", 0);
    markdown = g_string_new("# Search Results\n");
    g_string_append_printf(markdown, "Search terms: *%s*\n****\n", text);
    
    gtk_widget_set_sensitive(hv->window, FALSE);
    
    if ((dir = g_dir_open(hv->help_directory, 0, NULL))) {
        const gchar *name;
        
        while ((name = g_dir_read_name(dir))) {
#if GTK_CHECK_VERSION(2,16,0)
            gtk_entry_progress_pulse(GTK_ENTRY(hv->text_search));
#endif	/* GTK_CHECK_VERSION(2,16,0) */
            
            //if (g_str_has_suffix(name, ".md")) {
            if (TRUE) {
                FILE *file;
                gchar *path;
                gchar buffer[256];
                
                path = g_build_filename(hv->help_directory, name, NULL);
                if ((file = fopen(path, "rb"))) {
                    gboolean found = FALSE;
                    gchar *title = NULL;
                    
                    while (!found && fgets(buffer, sizeof buffer, file)) {
                        if (!title && (g_str_has_prefix(buffer, "# ") || g_str_has_prefix(buffer, " # "))) {
                            title = g_strstrip(strchr(buffer, '#') + 1);
                            title = g_strdup(title);
                        }
                        
                        for (term = 0; !found && terms[term]; term++) {
#ifdef strcasestr
                            found = strcasestr(buffer, terms[term]) != NULL;
#else
                            gchar *upper1, *upper2;
                            
                            upper1 = g_utf8_strup(buffer, -1);
                            upper2 = g_utf8_strup(terms[term], -1);
                            
                            found = strstr(upper1, upper2) != NULL;
                            
                            g_free(upper1);
                            g_free(upper2);
#endif
                        }
                    }
                    
                    if (found) {
                        no_results++;
                        
                        if (title) {
                              g_string_append_printf(markdown,
                                                     "* [%s](%s)\n", title, name);
                        } else {
                              g_string_append_printf(markdown,
                                                     "* [%s](%s)\n", name, name);
                        }
                    }
                    
                    g_free(title);
                    fclose(file);
                }
                
                g_free(path);
            }
        }
        
        g_dir_close(dir);
    }
    
    
    if (no_results == 0) {
        g_string_append_printf(markdown,
                               "Search returned no results.");
    } else {
        g_string_append_printf(markdown,
                               "****\n%d results found.", no_results);
    }
    
    /* shows the results inside the textview */
    markdown_textview_set_text(MARKDOWN_TEXTVIEW(hv->text_view), markdown->str);

    g_free(hv->current_file);
    hv->current_file = g_strdup_printf("search://%s", text);

#if GTK_CHECK_VERSION(2,16,0)
    gtk_entry_set_progress_fraction(GTK_ENTRY(hv->text_search), 0.0f);
#endif	/* GTK_CHECK_VERSION(2,16,0) */
    gtk_widget_set_sensitive(hv->window, TRUE);

    g_string_free(markdown, TRUE);
    g_strfreev(terms);
}

static void activate(GtkEntry *entry, gpointer data)
{
    HelpViewer *hv = (HelpViewer *)data;
    
    /* adds the current file to the back stack (before loading the new file */
    hv->back_stack = g_slist_prepend(hv->back_stack, g_strdup(hv->current_file));
    gtk_widget_set_sensitive(hv->btn_back, hv->can_go_back=TRUE);

    do_file_search((HelpViewer *)data, (gchar *)gtk_entry_get_text(entry));
}

#if GTK_CHECK_VERSION(2,16,0)
static void icon_press(GtkEntry *entry, gint position,
                       GdkEventButton *event, gpointer data)
{
    if (position == GTK_ENTRY_ICON_SECONDARY)
        activate(entry, data);
}
#endif	/* GTK_CHECK_VERSION(2,16,0) */ 

static void home_clicked(GtkWidget *button, gpointer data)
{
    HelpViewer *hv = (HelpViewer *)data;
	if (hv->homepage)
		help_viewer_open_page(hv, hv->homepage);
	else
		help_viewer_open_page(hv, DEFAULT_INDEX);
}

static gboolean home_key_pressed ( GtkAccelGroup  *group,
          GObject        *obj,
          guint           keyval,
          GdkModifierType mod,
          gpointer        data ) 
{
	home_clicked(NULL, data);
	return TRUE;
}

static void help_viewer_find_text(HelpViewer *hv) {
	static GtkWidget *dialog = NULL, *entry;
	GtkWidget *content_area, *label, *hbox, *error_dialog;
	int response, found;
		
	if (!dialog) {
		dialog = gtk_dialog_new_with_buttons (_("Find text"),
			GTK_WINDOW(hv->window), 0,
			GTK_STOCK_FIND, 1, _("Find _Previous"), 2, GTK_STOCK_CLOSE, GTK_RESPONSE_CLOSE, NULL);

		// create the widgets
		content_area = gtk_dialog_get_content_area (GTK_DIALOG (dialog));
		hbox = gtk_hbox_new (FALSE, 0);
		gtk_container_add (GTK_CONTAINER (content_area), hbox);
		
		label = gtk_label_new (_("Search for:"));
		gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 5);
		entry = gtk_entry_new ();
		gtk_box_pack_start (GTK_BOX (hbox), entry, TRUE, TRUE, 5);
		
		gtk_entry_set_activates_default (GTK_ENTRY(entry), TRUE);
		gtk_dialog_set_default_response	(GTK_DIALOG(dialog), 1);
	}
		
	gtk_widget_show_all (dialog);
	gtk_widget_grab_focus (entry);
	response = gtk_dialog_run (GTK_DIALOG(dialog));
	//g_print("response %d\n", response);
	gtk_widget_hide (dialog);
	
	if (response>0) { // stock responses are all negative
		const gchar * search_text = gtk_entry_get_text(GTK_ENTRY(entry));
		switch (response)
		{
			case 1: 
				found = markdown_textview_find_text (MARKDOWN_TEXTVIEW(hv->text_view), 
							search_text, MARKDOWN_TEXTVIEW_SEARCH_FORWARD);
				break;

			case 2: 
				found = markdown_textview_find_text (MARKDOWN_TEXTVIEW(hv->text_view), 
							search_text, MARKDOWN_TEXTVIEW_SEARCH_BACKWARD);
				break;

		}
		if (!found) 
		{
			error_dialog = gtk_message_dialog_new (GTK_WINDOW(hv->window),
							0, GTK_MESSAGE_ERROR, GTK_BUTTONS_CLOSE,
							_("\"%s\" not found."),search_text);
			gtk_dialog_run (GTK_DIALOG (error_dialog));
			gtk_widget_destroy (error_dialog);	
		}
	}
}

static void find_btn_clicked(GtkWidget *button, gpointer data)
{
    HelpViewer *hv = (HelpViewer *)data;
    help_viewer_find_text (hv);
}

static gboolean find_key_pressed ( GtkAccelGroup  *group,
          GObject        *obj,
          guint           keyval,
          GdkModifierType mod,
          gpointer        data ) 
{
	HelpViewer *hv = (HelpViewer *)data;
	help_viewer_find_text (hv);
	return TRUE;
}

void help_viewer_open_page(HelpViewer *hv, const gchar *page)
{
    gchar *temp;

    temp = g_strdup(hv->current_file);
    
    if (!markdown_textview_load_file(MARKDOWN_TEXTVIEW(hv->text_view), page)) {
        GtkWidget	*dialog;

        dialog = gtk_message_dialog_new(GTK_WINDOW(hv->parent),
                                        GTK_DIALOG_DESTROY_WITH_PARENT,
                                        GTK_MESSAGE_ERROR,
                                        GTK_BUTTONS_CLOSE,
                                        "Cannot open help file (%s).",
                                        page);
        gtk_dialog_run(GTK_DIALOG(dialog));
        gtk_widget_destroy(dialog);
        
        g_free(temp);
    } else {
        /* adds the current file to the back stack (before loading the new file */
        hv->back_stack = g_slist_prepend(hv->back_stack, temp);
        gtk_widget_set_sensitive(hv->btn_back, hv->can_go_back=TRUE);
        gtk_window_present(GTK_WINDOW(hv->window));
    }
}

void help_viewer_destroy(HelpViewer *hv)
{
    GSList *item;

    for (item = hv->back_stack; item; item = item->next) {
        g_free(item->data);
    }

    for (item = hv->forward_stack; item; item = item->next) {
        g_free(item->data);
    }
    
    g_slist_free(hv->back_stack);
    g_slist_free(hv->forward_stack);

    g_free(hv->current_file);
    g_free(hv->help_directory);

    if (!hv->parent) 
		gtk_main_quit();
}

static gboolean destroy_me(GtkWidget *widget, gpointer data)
{
    HelpViewer *hv = (HelpViewer *)data;
    
    help_viewer_destroy(hv);

    return FALSE;
}

HelpViewer *
help_viewer_new (const gchar *help_dir, const gchar *help_file, 
                 const gchar *title, GtkWindow *parent, 
                 gboolean enable_exec, gboolean use_gettext, gboolean use_extensions)
{
    HelpViewer *hv;
    GtkWidget *help_viewer;
    GtkWidget *vbox;
    GtkWidget *hbox;
    GtkWidget *toolbar1;
    GtkWidget *btn_back;
    GtkWidget *btn_forward;
    GtkWidget *separatortoolitem1;
    GtkWidget *toolbar2;
    GtkWidget *toolitem3;
#if !GTK_CHECK_VERSION(2,16,0)
    GtkWidget *label1;
#endif	/* GTK_CHECK_VERSION(2,16,0) */
    GtkWidget *toolitem4;
    GtkWidget *txt_search;
    GtkWidget *scrolledhelp_viewer;
    GtkWidget *markdown_textview;
    GtkWidget *status_bar;
    GtkWidget *btn_home, *btn_find;
    GdkPixbuf *icon;    
	GtkAccelGroup *accel;

    
    help_viewer = gtk_window_new(GTK_WINDOW_TOPLEVEL);
    gtk_widget_set_size_request(help_viewer, 300, 200);
    gtk_window_set_default_size(GTK_WINDOW(help_viewer), 640, 480);
    gtk_window_set_title(GTK_WINDOW(help_viewer), title?title:DEFAULT_WINDOW_TITLE);
    gtk_window_set_transient_for(GTK_WINDOW(help_viewer), parent);
    
    icon = gtk_widget_render_icon(help_viewer, GTK_STOCK_HELP, 
                                  GTK_ICON_SIZE_DIALOG,
                                  NULL);
    gtk_window_set_icon(GTK_WINDOW(help_viewer), icon);
    g_object_unref(icon);
    
    vbox = gtk_vbox_new (FALSE, 0);
    gtk_widget_show (vbox);
    gtk_container_add (GTK_CONTAINER (help_viewer), vbox);

    hbox = gtk_hbox_new (FALSE, 0);
    gtk_widget_show (hbox);
    gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);

    toolbar1 = gtk_toolbar_new ();
    gtk_widget_show (toolbar1);
    gtk_box_pack_start (GTK_BOX (hbox), toolbar1, TRUE, TRUE, 0);
    gtk_toolbar_set_style (GTK_TOOLBAR (toolbar1), GTK_TOOLBAR_BOTH_HORIZ);

    btn_back = (GtkWidget*) gtk_tool_button_new_from_stock ("gtk-go-back");
    gtk_widget_set_tooltip_text (btn_back, _("Backward in history (Alt-B)"));    
    gtk_widget_show (btn_back);
    gtk_container_add (GTK_CONTAINER (toolbar1), btn_back);
    gtk_tool_item_set_is_important (GTK_TOOL_ITEM (btn_back), TRUE);
    gtk_widget_set_sensitive(btn_back, FALSE);

    btn_forward = (GtkWidget*) gtk_tool_button_new_from_stock ("gtk-go-forward");
    gtk_widget_set_tooltip_text (btn_forward, _("Forward in history (Alt-F)"));    
    gtk_widget_show (btn_forward);
    gtk_container_add (GTK_CONTAINER (toolbar1), btn_forward);
    gtk_widget_set_sensitive(btn_forward, FALSE);

    separatortoolitem1 = (GtkWidget*) gtk_separator_tool_item_new ();
    gtk_widget_show (separatortoolitem1);
    gtk_container_add (GTK_CONTAINER (toolbar1), separatortoolitem1);

    btn_home = (GtkWidget*) gtk_tool_button_new_from_stock ("gtk-home");
    gtk_widget_set_tooltip_text (btn_home, _("Home (Alt-H)"));
    gtk_widget_show (btn_home);
    gtk_container_add (GTK_CONTAINER (toolbar1), btn_home);

    btn_find = (GtkWidget*) gtk_tool_button_new_from_stock ("gtk-find");
    gtk_widget_set_tooltip_text (btn_find, _("Find text (Ctrl-F)"));
    gtk_widget_show (btn_find);
    gtk_container_add (GTK_CONTAINER (toolbar1), btn_find);

    toolbar2 = gtk_toolbar_new ();
    gtk_widget_show (toolbar2);
    gtk_box_pack_end (GTK_BOX (hbox), toolbar2, FALSE, TRUE, 0);
    gtk_toolbar_set_style (GTK_TOOLBAR (toolbar2), GTK_TOOLBAR_BOTH_HORIZ);
    gtk_toolbar_set_show_arrow (GTK_TOOLBAR (toolbar2), FALSE);

    toolitem3 = (GtkWidget*) gtk_tool_item_new ();
    gtk_widget_show (toolitem3);
    gtk_container_add (GTK_CONTAINER (toolbar2), toolitem3);

#if !GTK_CHECK_VERSION(2,16,0)
    label1 = gtk_label_new_with_mnemonic ("_Search:");
    gtk_widget_show (label1);
    gtk_container_add (GTK_CONTAINER (toolitem3), label1);
#endif	/* GTK_CHECK_VERSION(2,16,0) */

    toolitem4 = (GtkWidget*) gtk_tool_item_new ();
    gtk_widget_show (toolitem4);
    gtk_container_add (GTK_CONTAINER (toolbar2), toolitem4);

    txt_search = gtk_entry_new ();
    gtk_widget_show (txt_search);
    gtk_container_add (GTK_CONTAINER (toolitem4), txt_search);
    gtk_entry_set_invisible_char (GTK_ENTRY (txt_search), 9679);
#if GTK_CHECK_VERSION(2,16,0)
    gtk_entry_set_icon_from_stock(GTK_ENTRY(txt_search),
                                  GTK_ENTRY_ICON_SECONDARY,
                                  GTK_STOCK_FIND);
#endif	/* GTK_CHECK_VERSION(2,16,0) */
	
    scrolledhelp_viewer = gtk_scrolled_window_new (NULL, NULL);
    gtk_widget_show (scrolledhelp_viewer);
    gtk_box_pack_start (GTK_BOX (vbox), scrolledhelp_viewer, TRUE, TRUE, 0);
    gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolledhelp_viewer), GTK_POLICY_NEVER, GTK_POLICY_AUTOMATIC);

    markdown_textview = markdown_textview_new();
    markdown_textview_set_image_directory(MARKDOWN_TEXTVIEW(markdown_textview), help_dir);
    markdown_textview_set_exec(MARKDOWN_TEXTVIEW(markdown_textview),enable_exec);
    markdown_textview_set_use_gettext(MARKDOWN_TEXTVIEW(markdown_textview),use_gettext);
    if (!use_extensions) markdown_textview_set_extensions(MARKDOWN_TEXTVIEW(markdown_textview), EGG_MARKDOWN_EXTENSION_NONE);
    gtk_widget_show (markdown_textview);
    gtk_container_add (GTK_CONTAINER (scrolledhelp_viewer), markdown_textview);

    status_bar = gtk_statusbar_new ();
    gtk_widget_show (status_bar);
    gtk_box_pack_start (GTK_BOX (vbox), status_bar, FALSE, FALSE, 0);

    hv = g_new0(HelpViewer, 1);
    hv->window = help_viewer;
    hv->status_bar = status_bar;
    hv->btn_back = btn_back;
    hv->btn_forward = btn_forward;
    hv->text_view = markdown_textview;
    hv->text_search = txt_search;
    hv->help_directory = g_strdup(help_dir);
    hv->back_stack = NULL;
    hv->forward_stack = NULL;
    hv->parent = GTK_WIDGET(parent);
    hv->can_go_back = hv->can_go_forward = FALSE;

    g_signal_connect(markdown_textview, "link-clicked", G_CALLBACK(link_clicked), hv);
    g_signal_connect(markdown_textview, "hovering-over-link", G_CALLBACK(hovering_over_link), hv);
    g_signal_connect(markdown_textview, "hovering-over-text", G_CALLBACK(hovering_over_text), hv);
    g_signal_connect(markdown_textview, "file-load-complete", G_CALLBACK(file_load_complete), hv);

    g_signal_connect(btn_back, "clicked", G_CALLBACK(back_clicked), hv);
    g_signal_connect(btn_forward, "clicked", G_CALLBACK(forward_clicked), hv);
    g_signal_connect(btn_home, "clicked", G_CALLBACK(home_clicked), hv);
    g_signal_connect(btn_find, "clicked", G_CALLBACK(find_btn_clicked), hv);

    g_signal_connect(help_viewer, "delete-event", G_CALLBACK(destroy_me), hv);
    g_signal_connect(txt_search, "activate", G_CALLBACK(activate), hv);

#if GTK_CHECK_VERSION(2,16,0)
    g_signal_connect(txt_search, "icon-press", G_CALLBACK(icon_press), hv);
#endif	/* GTK_CHECK_VERSION(2,16,0) */

	// attach Ctrl-F search key
	accel = gtk_accel_group_new();
	gtk_accel_group_connect (accel, gdk_keyval_from_name ("f"), GDK_CONTROL_MASK, 0, 
		g_cclosure_new (G_CALLBACK(find_key_pressed), hv, NULL));
	gtk_accel_group_connect (accel, gdk_keyval_from_name ("f"), GDK_MOD1_MASK, 0, 
		g_cclosure_new (G_CALLBACK(forward_key_pressed), hv, NULL));
	gtk_accel_group_connect (accel, gdk_keyval_from_name ("b"), GDK_MOD1_MASK, 0, 
		g_cclosure_new (G_CALLBACK(backward_key_pressed), hv, NULL));
	gtk_accel_group_connect (accel, gdk_keyval_from_name ("h"), GDK_MOD1_MASK, 0, 
		g_cclosure_new (G_CALLBACK(home_key_pressed), hv, NULL));		
	gtk_window_add_accel_group (GTK_WINDOW(help_viewer), accel);
                            
    if (!markdown_textview_load_file(MARKDOWN_TEXTVIEW(markdown_textview), help_file ? help_file : DEFAULT_INDEX)) {
        GtkWidget	*dialog;

        dialog = gtk_message_dialog_new(parent,
                                        GTK_DIALOG_DESTROY_WITH_PARENT,
                                        GTK_MESSAGE_ERROR,
                                        GTK_BUTTONS_CLOSE,
                                        "Cannot open help file (%s).",
                                        help_file ? help_file : DEFAULT_INDEX);
        gtk_dialog_run(GTK_DIALOG(dialog));
        gtk_widget_destroy(dialog);
        
        gtk_widget_destroy(hv->window);
        g_free(hv);
        
        return NULL;
    }

    gtk_widget_show_all (hv->window);
    gtk_widget_grab_focus (hv->text_view);

    return hv;
}

////============== stdout ================
void stdout_output (gchar *dir, gchar *file, gboolean enable_exec, gboolean use_gettext,
					int output_type, FILE *pot_output, gboolean use_extensions)
{
    int fd;
    gchar *path, *parsed=NULL;

    /* we do assume UTF-8 encoding */
    path = g_build_filename(dir, file?file:DEFAULT_INDEX, NULL);
    
	EggMarkdown *markdown = egg_markdown_new();    
	egg_markdown_set_output(markdown, output_type);	
    egg_markdown_set_autocode(markdown, TRUE);
    egg_markdown_set_use_gettext(markdown, use_gettext);
    if (!use_extensions) egg_markdown_set_extensions(markdown, EGG_MARKDOWN_EXTENSION_NONE);
	if (output_type == EGG_MARKDOWN_OUTPUT_HTML)
	    egg_markdown_set_smart_quoting(markdown, TRUE);
	if (enable_exec) 
		egg_markdown_set_exec (markdown, enable_exec);
	if (pot_output) 
		egg_markdown_set_pot_output (markdown, pot_output, realpath(path, NULL));	
    
    if ((fd=open(path,O_RDONLY))>=0) {
		struct stat sb;
		fstat(fd, &sb);

		gchar *buffer = mmap(NULL, sb.st_size, PROT_READ, MAP_SHARED | MAP_NORESERVE, fd, 0);
		if (buffer!=(gchar *)-1) {
			egg_markdown_clear(markdown);
			parsed = egg_markdown_parse(markdown, buffer);
			munmap(buffer, sb.st_size);

			if (!pot_output) {
				write(1, parsed, strlen(parsed));
				write(1, "\n", 1);
			}
			g_free(parsed);
		}
	}
	close(fd);
	g_free(path);

}

void usage(int exitcode) {
	g_print(MDVIEW_VERSION_TEXT);	
	g_print("\
Usage: mdview [options] [path [home [title]]]\n\
 - path - which directory/file to open. All links are relative to path given.\n\
   If path is directory, " DEFAULT_INDEX " will be opened.\n\
 - home file. if not specified, " DEFAULT_INDEX " is used.\n\
 - title - window title, if not specified, " DEFAULT_WINDOW_TITLE " is used\n\
You can skip parameters by specifying blanks. \n\n\
Options:\n\
 --exec or -x : runs executable code\n\
 --tty        : output to stdout as text with vt100 codes\n\
 --text       : output to stdout as plain text\n\
 --html       : output to stdout as html fragment\n\
 --po         : output to stdout as .po file (for translation)\n\
 --no-gettext : don\'t use internal gettext\n\
 --no-ext     : disable all extensions\n\
 \n\
"	);
	exit(0);
}
////================ main ================

int main(int argc, char **argv)
{
    HelpViewer *hv;
    gchar *startup_file=NULL;
    gchar *file=NULL;
    gchar *home=NULL;
    gchar *dir=NULL;
    gchar *title=NULL;
    gboolean enable_exec = FALSE;
    gboolean console_output = FALSE;
    gboolean use_gettext = TRUE;
    gboolean use_extensions = TRUE;
    int output_type = EGG_MARKDOWN_OUTPUT_TTY;
    FILE *pot_output = NULL;
    int i;
    
    // locale init
    gchar *temp;    
	setlocale( LC_ALL, "" );   
    if (temp=getenv("TEXTDOMAIN")) textdomain(temp);
    if (temp=getenv("TEXTDOMAINDIR")) bindtextdomain(textdomain(NULL), temp);
    //g_print("%s\n",textdomain(NULL));
    //g_print("%s\n", bindtextdomain(textdomain(NULL), NULL ));
    //exit(0);

	for (i=1; i<argc; i++) {
		if (strcmp(argv[i],"-h")==0 ||
		    strcmp(argv[i],"--help")==0)
		{
			usage(0);
		} else 	if (strcmp(argv[i],"-V")==0 ||
		    strcmp(argv[i],"--version")==0)
		{
			g_print(MDVIEW_VERSION_TEXT);
			exit(0);
		} else if (strcmp(argv[i],"-x")==0 ||
		    strcmp(argv[i],"--exec")==0)
		{
			enable_exec = TRUE;
			continue;
		} else if (strcmp(argv[i],"--tty")==0) {
			console_output = TRUE;
			output_type = EGG_MARKDOWN_OUTPUT_TTY;
			continue;
		} else if (strcmp(argv[i],"--text")==0) {
			console_output = TRUE;
			output_type = EGG_MARKDOWN_OUTPUT_TEXT;
			continue;
		} else if (strcmp(argv[i],"--html")==0) {
			console_output = TRUE;
			output_type = EGG_MARKDOWN_OUTPUT_HTML;
			continue;
		} else if (strcmp(argv[i],"--po")==0) {
			console_output = TRUE;
			pot_output = stdout;
			continue;
		} else if (strcmp(argv[i],"--no-gettext")==0) {
			use_gettext = FALSE;
			continue;
		} else if (strcmp(argv[i],"--no-ext")==0) {
			use_extensions = FALSE;
			continue;			
		} else if (!startup_file) {
			startup_file=argv[i];
		} else if (!home) {
			home=argv[i];
		} else if (!title) {
			title=argv[i];			
		}
	}
	
	
	// defaults
	if (startup_file && g_file_test (startup_file, G_FILE_TEST_EXISTS)) {
		if (g_file_test (startup_file, G_FILE_TEST_IS_DIR)) {
			dir = startup_file;
			file = DEFAULT_INDEX;
		} else {
			dir = g_path_get_dirname(startup_file);
			file = g_path_get_basename(startup_file);
		}
	} else {
		usage(1);
	}
    if (home && !home[0]) home=NULL;
    if (title && !title[0]) title=NULL;

	if (console_output || !getenv("DISPLAY")) {
		// output to stdout 
		stdout_output (dir, file, enable_exec, use_gettext, output_type, pot_output, use_extensions);
	} else {
		// run GUI
		gtk_init(&argc, &argv);
		hv = help_viewer_new(dir, file, title, NULL, enable_exec, use_gettext, use_extensions);
		if (hv==NULL) exit(1);
		hv->homepage = home;
		
		gtk_main();
	}
}

